# -*- coding: utf-8 -*-
###########################################################################
#
# Copyright (C) 2010  Bartek thindil Jasicki
#
# This file is part of Anaria MUD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
'''
account_handler.py

The login and creation of accounts, and handles all account procedures, loading
and deleting, of characters.
'''
import mud, mudsys, char, hooks, account, event, telnetlib, utils, os, time, mudsock

# MCCP
MCCP2 = 'V'
MCCP = 'U'

# TTYPE
TTYPE = '\x18'

# MXP
MXP = '\x5B'

# handshake crap
server_will  = telnetlib.IAC + telnetlib.WILL
server_wont  = telnetlib.IAC + telnetlib.WONT
client_do    = telnetlib.IAC + telnetlib.DO
client_dont  = telnetlib.IAC + telnetlib.DONT

# subnegotiation crap
subneg_start_ttype = telnetlib.IAC + telnetlib.SB   + TTYPE
subneg_end   = telnetlib.IAC + telnetlib.SE

# control sequences for squelching passwords
squelch   = server_will + telnetlib.ECHO
unsquelch = server_wont + telnetlib.ECHO

def check_acct_name(name):
    '''Makes sure an account name is valid'''
    return (len(name) > 3 and len(name) < 13 and
            name[0].isalpha() and name.isalnum())

def login_method_prompt(sock):
    sock.send_raw("Wybierz opcje: ")

def acct_new_password_prompt(sock):
    sock.send_raw("\r\nJakie będzie twoje nowe hasło? " + squelch)

def acct_password_prompt(sock):
    sock.send_raw("Jakie jest twoje hasło? " + squelch)

def acct_wait_dns_prompt(sock):
    sock.send_raw("\r\n")

def try_create_account(sock, name, psswd):
    if mudsys.account_exists(name):
        return False
    elif mudsys.account_creating(name):
        return False
    elif not check_acct_name(name):
        return False
    else:
        # creating a new account
        mud.log_string("Account '" + name + "' is trying to create.")

        # create our new account
        acct = mudsys.create_account(name)
        if acct == None:
            return False
        else:
            mudsys.attach_account_socket(acct, sock)
            mudsys.set_password(acct, psswd)
            sock.pop_ih()
            sock.push_ih(acct_menu_handler, acct_main_menu)

            #sock.push_ih(acct_finish_handler, acct_finish_prompt)

            # log that the account created
            mud.log_string("New account '" + acct.name + "' has created.")

            # register and save the account to disk
            mudsys.do_register(acct)

            return True
    return False

def try_load_account(sock, name, psswd):
    '''Attempt to load an account with the given name and password.'''
    if not mudsys.account_exists(name):
        return False
    else:
        acct = mudsys.load_account(name)
        if not mudsys.password_matches(acct, psswd):
            return False

        # successful load
        mudsys.attach_account_socket(acct, sock)
        sock.pop_ih()
        sock.push_ih(acct_menu_handler, acct_main_menu)
        return True
    return False

def login_method_handler(sock, arg):
    args = arg.split()
    if len(args) == 0:
        return
    args[0] = args[0].lower()

    if "rejestruj".startswith(args[0]):
        if len(args) != 3:
            return
        elif mudsys.account_exists(args[1]):
            sock.send("{cKonto o tej nazwie już istnieje.{n\r\n")
        elif not check_acct_name(args[1]):
            txt = "{cTo jest nieprawidłowa nazwa konta. Nazwa konta musi składać się "\
                  "tylko z liter oraz cyfr i musi mieć od 4 do 12 znaków długości. " \
                  "Pierwszy znak MUSI być literą. Proszę wybrać inną nazwę konta.{n"
            sock.send(mud.format_string(txt, False))

        elif not try_create_account(sock, args[1], args[2]):
            sock.send("Nie można utworzyć twojego konta.")

    elif "loguj".startswith(args[0]):
        if len(args) != 3:
            return
        elif not try_load_account(sock, args[1], args[2]):
            sock.send("{cNieprawidłowa nazwa konta bądź hasło.{n\r\n")
    elif "gosc".startswith(args[0]):
        if (mudsys.sys_getval("lockdown") != '' and
            not utils.is_keyword(mudsys.sys_getval("lockdown"), "player")):
            sock.send("{cGra jest obecnie zamknięta dla nowych graczy.{n\r\n")
        else:
            sock.pop_ih()
            hooks.run("create_guest", hooks.build_info("sk", (sock,)))
    elif "koniec".startswith(args[0]):
        sock.close()

def acct_wait_dns_handler(sock, arg):
    # do nothing
    return

def acct_new_password_handler(sock, arg):
    '''asks a new account for a password'''
    sock.send_raw(unsquelch)
    if len(arg) > 0:
        # put in mudsys to prevent scripts from messing with passwords
        mudsys.set_password(sock.account, arg)
        sock.pop_ih()

def acct_password_handler(sock, arg):
    '''asks an account to verify its password'''
    # password functions put in mudsys to prevent scripts from
    # messing with passwords
    sock.send_raw(unsquelch)
    if not mudsys.password_matches(sock.account, arg):
        sock.send("Nieprawidłowe hasło.")
        sock.close()
    else:
        # password matches, pop our handler and go down a level
        sock.pop_ih()

def acct_encoding_handler(sock, arg):
    '''set character encoding for account'''
    args = arg.split()
    if args[0].isdigit():
        encoding = ['bez zmian', 'utf-8', 'windows-1250', 'iso-8859-2', 'brak']
        opt  = int(args[0])
        if opt < 0 or opt >= len(encoding):
            sock.send("Nieprawidłowy wybór!")
        else:
            sock.account.aux('data').encoding = encoding[opt]
            mudsys.do_save(sock.account)
            sock.send("Kodowanie znaków zmieniona na: " + encoding[opt])
            sock.pop_ih()
    elif args[0].upper() == 'Q':
        sock.pop_ih()
    else:
        sock.send("Nieprawidłowy wybór!")

def acct_encoding_menu(sock):
    '''show available encodings to player.'''
    buf = ['Zmień te ustawienia tylko wtedy, kiedy nie widzisz polskich znaków, bądź zamiast nich na ekranie masz dziwne znaki. Dostępne opcje:']
    encoding = ['bez zmian', 'utf-8', 'windows-1250', 'iso-8859-2', 'brak']
    i = 0
    for enc in encoding:
        if enc == sock.account.aux('data').encoding:
            buf.append('%d)%s (obecne)' % (i, enc))
        else:
            buf.append('%d)%s' % (i, enc))
        i += 1
    buf.append('Wybierz opcję lub Q aby wyjść:')
    for text in buf:
        sock.send(text)

def find_reconnect(name):
    '''searches through the character list for a PC whose name matches the
       supplied name'''
    name = name.lower()
    for ch in char.char_list():
        if ch.is_pc and name == ch.name.lower():
            return ch
    return None

def acct_load_char(sock, name):
    '''loads a character attached to the account. Argument supplied must be a
       name of the corresponding character'''
    chars = sock.account.characters()
    
    if not name.lower() in [n.lower() for n in sock.account.characters()]:
        sock.send("Postać o tym imieniu nie istnieje na twoim koncie.")
    else:
        # first, try a reconnect
        ch = find_reconnect(name)

        # no reconnect? Try loading our character
        if not ch == None:
            # swap the sockets
            old_sock = ch.sock

            # attach and detach put in mudsys to prevent scripts from
            # messing around with the connections between chars and sockets
            mudsys.detach_char_socket(ch)
            mudsys.attach_char_socket(ch, sock)
            
            if old_sock != None:
                old_sock.close()
            mud.log_string(ch.name + " has reconnected.")
            ch.send("Przejmujesz kontrolę nad używanym ciałem.")
            ch.act("poczta")
            ch.act("patrz")
            hooks.run("reconnect", hooks.build_info("ch", (ch,)))
            sock.push_ih(mudsys.handle_cmd_input, mudsys.show_prompt, "playing")

        else:
            # load our character. Put in mudsys to prevent scripts from using it
            ch = mudsys.load_char(name)
            if ch == None:
                sock.send("Plik gracza dla " + name + " nie istnieje.")
            elif (mudsys.sys_getval("lockdown") != '' and
                  not ch.isInGroup(mudsys.sys_getval("lockdown"))):
                sock.send("Nie możesz obecnie wejść do gry.")
                mud.extract(ch)
            elif ch.isInGroup('banned'):
                sock.send('Ta postać została zbanowana.')
                mud.extract(ch)
            else:
                # attach put in mudsys to prevent scripts from messing with
                # character and socket links
                mudsys.attach_char_socket(ch, sock)
                if mudsys.try_enter_game(ch):
                    mud.log_string(ch.name + " has entered the game.")
                    sock.push_ih(mudsys.handle_cmd_input, mudsys.show_prompt,
                                 "playing")
                    ch.act("poczta")
                    ch.act("patrz")
                    hooks.run("enter", hooks.build_info("ch rm", (ch, ch.room)))
                else:
                    sock.send("Wystąpił problem podczas wchodzenia do gry. " + \
                              "Spróbuj ponownie później.")
                    # detach put in mudsys t prevent scripts from messing with
                    # character and socket links
                    mudsys.detach_char_socket(ch, sock)
                    mud.extract(ch)

def acct_menu_handler(sock, arg):
    '''parses account commands (new character, enter game, quit, etc)'''
    if len(arg) == 0:
        return

    args = arg.split()
    args[0] = args[0].upper()
    
    if args[0].isdigit():
        opts = sock.account.characters()
        arg  = int(args[0])
        if arg < 0 or arg >= len(opts):
            sock.send("Nieprawidłowy wybór!")
        else:
            acct_load_char(sock, opts[arg])
    elif args[0] == 'Q' or "QUIT".startswith(args[0]):
        sock.send("Wróć wkrótce!")
        mudsys.do_save(sock.account)
        sock.close()
    elif args[0] == 'Z':
        # sock.push_ih(acct_confirm_password_handler,acct_confirm_password_prompt)
        sock.push_ih(acct_new_password_handler, acct_new_password_prompt)
        sock.push_ih(acct_password_handler, acct_password_prompt)
    elif args[0] == 'N':
        if "player" in [x.strip() for x in mudsys.sys_getval("lockdown").split(",")]:
            sock.send("Nie można na razie tworzyć nowych postaci.")
        else:
            hooks.run("create_character", hooks.build_info("sk", (sock,)))
    elif args[0] == 'U':
        sock.push_ih(acct_encoding_handler, acct_encoding_menu)
    elif args[0] == "SKASUJ":
        if len(args) < 2:
            sock.send("Skasować co?")
        elif args[1].upper() != "KONTO":
            chars = sock.account.characters()
            chars.sort()
            if len(chars) > 0:
                for i in range(len(chars)):
                    if args[1].lower() == chars[i].lower():
                        os.unlink("../lib/players/pfiles/" + chars[i][0] + "/" + chars[i] + ".pfile")
                        os.unlink("../lib/players/objfiles/" + chars[i][0] + "/" + chars[i] + ".ofile")
                        sock.account.remove_char(args[1].lower())
                        mudsys.do_save(sock.account)
                        sock.send("Postać %s została skasowana." % chars[i])
                        break
        elif args[1].upper() == "KONTO":
            chars = sock.account.characters()
            if len(chars) > 0:
                for char in chars:
                    os.unlink("../lib/players/pfiles/" + char[0] + "/" + char + ".pfile")
                    os.unlink("../lib/players/objfiles/" + char[0] + "/" + char + ".ofile")
            os.unlink("../lib/accounts/" + sock.account.name[0].upper() + "/" + sock.account.name[0].upper()
                      + sock.account.name[1:] + ".acct")
            sock.close()
    else:
        sock.send("Nieprawidłowy wybór!" + args)

def acct_main_menu(sock):
    '''displays the main menu for the account and asks for a command'''
    line_buf = "%-38s" % " "

    sock.account.aux('data').lastvisited = time.time()
    if sock.account.aux('data').encoding != 'bez zmian':
        sock.aux('misc').encoding = sock.account.aux('data').encoding

    # make the account menu look pretty
    img = ["+--------------------------------------+",
           "|        |~        jgs                 |",
           "|       /.\       /#^^\_     /\        |",
           "|      /#  \     /#%    \   /# \       |",
           "|     /#%   \   /#%______\ /#%__\      |",
           "|    /#%     \   |= I I ||  |- |       |",
           "|    ~~ ~~~|~~   |_=_-__|'  |[]|       |",
           "|      |[] |_______\__|/_ _ |= |`.     |",
           "|      |-  /= __   __    /-\|= | :;    |",
           "|      |= /- /\/  /\/   /=- \.-' :;    |",
           "|      | /_.=========._/_.-._\  .:'    |",
           "|      |= |-.'.- .'.- |  /|\ |.:'      |",
           "|      \  |=|:|= |:| =| |~|~||'|       |",
           "|       |~|-|:| -|:|  |-|~|~||=|       |",
           "|       |=|=|:|- |:|- | |~|~|| |       |",
           "|       | |-_~__=_~__=|_^^^^^|/___     |",
           "|       |-(=-=-=-=-=-(|=====/=_-=/\    |",
           "|       | |=_-= _=- _=| -_=/=_-_/__\   |",
           "|       | |- _ =_-  _-|=_- |]#| I II   |",
           "|       |=|_/ \_-_= - |- = |]#| I II   |",
           "|       | /  _/ \. -_=| =__|]!!!I_II!! |",
           "|      _|/-'/  ` \_/ \|/' _ ^^^^`.==_^.|",
           "+--------------------------------------+"]

    # the lines for displaying our account options
    opts  = [ ]
    chars = sock.account.characters()
    chars.sort()
    if len(chars) > 0:
        opts.append("  {wGraj Postacią:                      ")
        for i in range(len(chars)):
            opts.append("    {n[{c%1d{n] %-30s" % (i,chars[i]))
        opts.append(line_buf)

    # append our other displays
    opts.append("  {wZarządzanie Kontem:                 ")
    opts.append("    {n[{cZ{n]miana hasła                    ")
    opts.append("    {n[{cN{n]owa postać                     ")
    opts.append("    {n[{cU{n]staw kodowanie znaków          ")
    opts.append("    {n[{cSkasuj{n] postać [{cimię postaci{n]    ")
    opts.append("    {n[{cSkasuj konto{n]                    ")
    opts.append(line_buf)
    opts.append(line_buf)
    opts.append(line_buf)

    # fill up our height to be in line with the image
    while len(opts) < len(img) - 2:
        opts.insert(0, line_buf)

    # append our title
    opts.insert(1, "            {nW I T A J  W              ")
    opts.insert(2, "             {nA N A R I I              ")
            
    # display all of our info
    sock.send("")
    for i in range(max(len(opts), len(img))):
        if i < len(opts):
            sock.send_raw(opts[i])
        else:
            sock.send_raw(line_buf)
        if i < len(img):
            sock.send_raw("{n%s" % img[i])
        sock.send("")
        
    sock.send_raw("{nWpisz opcję lub Q aby wyjść: ")



################################################################################
# events for blocking action when dns lookup is in progress
################################################################################
def dns_check_event(owner, void, info):
    '''After a socket connects, monitor their hostname until dns lookup is
       complete. Then, put the socket into the account handler.
    '''
    sock, = hooks.parse_info(info)
    if sock != None and sock.can_use:
        mud.log_string("New connection from " + sock.hostname)
        sock.send_raw(server_will + MCCP2)
        sock.send_raw(client_do + TTYPE)
        sock.send_raw(server_will + MXP)
        sock.send(mud.get_greeting())
        sock.push_ih(login_method_handler, login_method_prompt)
        sock.send("== Opcje ======================================================================")
        sock.send("    Zaloguj sie     : loguj     <konto> <haslo>")
        sock.send("    Rejestruj sie   : rejestruj <konto> <haslo>")
        sock.send("    Graj jako gosc  : gosc")
        sock.send("    Wyjdz           : koniec")
        sock.send("===============================================================================")
        sock.send("")
        sock.bust_prompt()
    else:
        event.start_event(None, 0.2, dns_check_event, None, info)



################################################################################
# account handler hooks
################################################################################
def account_handler_hook(info):
    '''When a socket connects, put them into the account handler menu.'''
    # put a nonfunctional prompt up while waiting for the DNS to resolve
    sock, = hooks.parse_info(info)
    sock.push_ih(acct_wait_dns_handler, acct_wait_dns_prompt)
    event.start_event(None, 0.2, dns_check_event, None, info)


def copyover_complete_hook(info):
    sock, = hooks.parse_info(info)
    sock.send_raw(server_will + MCCP2)
    sock.send_raw(server_will + MXP)
    sock.send_raw(client_do + TTYPE)
    sock.push_ih(acct_menu_handler, acct_main_menu)
    sock.push_ih(mudsys.handle_cmd_input, mudsys.show_prompt, "playing")

def iac_hook(info):
    '''Check for IAC commands.'''
    sock, cmd = hooks.parse_info(info)
    if cmd == client_do + MCCP2:
        mud.log_string("MCCP2 support detected on " + sock.hostname)
        sock.start_compress("MCCP2")
    elif cmd == client_dont + MCCP2:
        sock.send_raw(server_will + MCCP)
    elif cmd == client_do + MCCP:
        mud.log_string("MCCP support detected on " + sock.hostname)
        sock.start_compress("MCCP")
    elif cmd == server_will + TTYPE:
        sock.send_raw(subneg_start_ttype + '\x01' + subneg_end)
    elif cmd == server_wont + TTYPE:
        mud.log_string("Unknown client type on " + sock.hostname)
    elif cmd.startswith(subneg_start_ttype + 'F'):
        client = cmd[4:-2]
        sock.aux('misc').encoding = 'brak'
        mud.log_string("Client type: %s on %s." % (client, sock.hostname))
        clients = { 'utf-8' : ['KildClient', 'TinTin++', 'hmud'],
                    'windows-1250' : ['mushclient']
                    }
        for key in clients.keys():
            if client in clients[key]:
                sock.aux('misc').encoding = key
                break
    elif cmd == client_do + MXP:
        mud.log_string("MXP support detected on " + sock.hostname)
        sock.aux('misc').mxp = True
        sock.send_raw(telnetlib.IAC + telnetlib.SB + MXP + '\x01' + subneg_end)

################################################################################
# loading and unloading the module
################################################################################
hooks.add("receive_connection", account_handler_hook)
hooks.add("copyover_complete",  copyover_complete_hook)
hooks.add("receive_iac", iac_hook)

def __unload__():
    '''removes the hooks for account handling'''
    hooks.remove("receive_connection", account_handler_hook)
    hooks.remove("copyover_complete",  copyover_complete_hook)
    hooks.remove("receive_iac", iac_hook)
