# -*- coding: utf-8 -*-
###########################################################################
#
# Copyright (C) 2010  Bartek thindil Jasicki
#
# This file is part of Anaria.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
'''
armors.py

Armors item type.
'''
import olc, mudsys, auxiliary, mudsock, storage, mud, utils, hooks



################################################################################
# item type definition
################################################################################

# a map from our type to possible subtypes.
# Also the default subtype when we create an item of the specified type.
atype_map  = { }
atype_dflt = { }

class ArmorData:
    '''Data for armors - type, subtype and resistance'''

    __item_type__ = "armor"

    def __init__(self, set = None):
        self.type = "armor"
        self.subtype = "skin"
        self.resistance = 1

        if set != None:
            self.type = set.readString('type')
            self.subtype = set.readString('subtype')
            self.resistance = set.readInt('resistance')

    def copyTo(self, to):
        to.type = self.type
        to.subtype = self.type
        to.resistance = self.resitance

    def copy(self):
        newData = ArmorData()
        self.copyTo(newData)
        return newData

    def store(self):
        set = storage.StorageSet()
        set.storeString('type', self.type)
        set.storeString('subtype', self.subtype)
        set.storeInt('resistance', self.resistance)
        return set

################################################################################
# creating and interacting with armor types
################################################################################
def register_armor_type(type, dflt_subtype = "unknown"):
    '''create a new armor type, and add a default subtype'''
    subtypes = [dflt_subtype]
    atype_map[type]  = subtypes
    atype_dflt[type] = dflt_subtype

def register_armor_subtype(type, subtype):
    '''create a new subtype for the specified armor type'''
    if atype_map.has_key(type):
        subtype  = subtype.lower()
        subtypes = atype_map[type]
        if not subtype in subtypes:
            subtypes.append(subtype)

def get_armor_dflt(type):
    '''get the default subtype for a armor'''
    if atype_dflt.has_key(type):
        return atype_dflt[type]
    return None

def get_armor_types():
    '''list all armors types'''
    return atype_map.keys()

def get_armor_subtypes(type):
    '''list all subtypes for a armor'''
    if atype_map.has_key(type):
        return [t for t in atype_map[type]]
    return None

def is_armor_type(type):
    '''return whether type is a registered armor type'''
    return atype_map.has_key(type)

def is_armor_subtype(type, subtype):
    '''return whether subtype is a registered subtype for type'''
    if atype_map.has_key(type):
        return subtype in atype_map[type]
    return False
    


################################################################################
# Armor OLC
################################################################################
__ARMOLC_TYPE__         = 1
__ARMOLC_SUBTYPE__      = 2
__ARMOLC_RESISTANCE__   = 3

armolc_opt_map = {
    '1' : (__ARMOLC_TYPE__,     "Wybierz typ: "),
    '2' : (__ARMOLC_SUBTYPE__,  "Wybierz podtyp: "),
    '3' : (__ARMOLC_RESISTANCE__, "Wprowadź obronę: "),
    }

def armolc_menu(sock, data):
    '''display our menu interface'''
    sock.send("{g1) Typ         : {c" + data.type    + "\n" + \
              "{g2) Podtyp      : {c" + data.subtype + "\n" + \
              "{g3) Obrona      : {c" + str(data.resistance) + "\n")

def armolc_chooser(sock, data, option):
    '''figure out what field we want to edit, and display a prompt'''
    if not armolc_opt_map.has_key(option):
        return olc.MENU_CHOICE_INVALID
    ret, mssg = armolc_opt_map[option]

    # list our type and subtype options
    opts = None
    if ret == __ARMOLC_TYPE__:
        opts = get_armor_types()
    elif ret == __ARMOLC_SUBTYPE__:
        opts = get_armor_subtypes(data.type)

    # do we have to list subtype options?
    if opts != None:
        utils.olc_display_table(sock, opts, 4)
    sock.send_raw(mssg)
    return ret

def armolc_parser(sock, data, choice, arg):
    '''parse out what we want to change a specified field to'''
    try:
        arg = int(arg)
    except: return False
    
    if choice == __ARMOLC_TYPE__:
        opts = get_armor_types()
        arg  = int(arg)
        if arg < 0 or arg >= len(opts):
            return False

        newtype = opts[arg]
        if newtype != data.type:
            data.resistance = 1
            data.subtype = get_armor_dflt(newtype)
            data.type = newtype

    elif choice == __ARMOLC_SUBTYPE__:
        opts = get_armor_subtypes(data.type)
        arg  = int(arg)
        if arg < 0 or arg >= len(opts):
            return False
        data.subtype = opts[arg]

    elif choice == __ARMOLC_RESISTANCE__:
        data.resistance = max(int(arg), 1)
    else:
        return False
    return True

def armolc_to_proto(data):
    '''return a script that will generate relevant info to create this item'''
    buf  = 'me.get_type_data("armor").type = "' + data.type + '"\n'
    buf += 'me.get_type_data("armor").subtype = "' + data.subtype + '"\n'
    buf += 'me.get_type_data("armor").resistance = ' + str(data.resistance) + '\n'
    return buf

################################################################################
# game hooks
################################################################################
def armor_desc_info(info):
    '''appends armor information about an object to the description
       buffer when a person looks at it.'''
    obj, ch = hooks.parse_info(info)

    # no point in continuing
    if not obj.istype("armor"):
        return

    # it is armor or shield?
    if obj.get_type_data("armor").type == "armor":
        desc = " Założone, obniża otrzymywane obrażenia na danej części ciała o " + str(obj.get_type_data("armor").resistance) + "%."
    else:
        desc = " Założona, podnosi szansę zablokowania ciosu o " + str(obj.get_type_data("armor").resistance) + "."

    # append our description
    ch.look_buf += desc

################################################################################
# initialization
################################################################################

# initialize our item data, and its OLC
mudsys.item_add_type(ArmorData.__item_type__, ArmorData)
olc.item_add_olc(ArmorData.__item_type__,
                 armolc_menu, armolc_chooser, armolc_parser,
                 None, armolc_to_proto)

# display hooks for armors
hooks.add("append_obj_desc", armor_desc_info)

# initialize some default armor types and subtypes
register_armor_type("armor",    "skin")
register_armor_subtype("armor", "chain")
register_armor_subtype("armor", "plate")
register_armor_type("shield", "small")
register_armor_subtype("shield", "medium")
register_armor_subtype("shield", "large")

# add worn type for armors
mudsys.add_worn_type('shield', 'trzymane')
mudsys.add_worn_type('hełm', 'głowa')
mudsys.add_worn_type('spodnie', 'noga, noga')
mudsys.add_worn_type('buty', 'stopa, stopa')
mudsys.add_worn_type('rękawice', 'dłoń, dłoń')
mudsys.add_worn_type('naszyjnik', 'szyja')
mudsys.add_worn_type('pierścień', 'palec')
mudsys.add_worn_type('pas', 'pas')
mudsys.add_worn_type('maska', 'twarz')
mudsys.add_worn_type('pełny hełm', 'głowa, twarz')
mudsys.add_worn_type('kolczyk', 'ucho')
mudsys.add_worn_type('backpack', 'plecy')
mudsys.add_worn_type('bag', 'przy ciele')
mudsys.add_worn_type('szata', 'tułów, ręka, ręka, noga, noga')

def __unload__():
    '''remove all hooks.'''
    hooks.remove('append_obj_desc', armor_desc_info)
