# -*- coding: utf-8 -*-
################################################################################
#
# bulletin.py
#
# This module implements a bulletin board system, allowing people to post and
# read messages to various bulletins. This is *not* a substitute for forums. It
# is not nearly as functionally complete as any freely available web-based
# forum. This is meant for posting notes and notices, not for carrying out
# conversations.
#
# contains one command, "bulletin". Syntax is:
#   bulletin
#   bulletin read <board> [article number]
#   bulletin post <board> <title>
#   bulletin delete <board> [article number]
#   bulletin create <board> [groups]
#
# Post makes use of a character's notepad, from the editor module.
#
# Bulletin requires a minimum of NakedMud v3.5 to work, but should work fine
# with version below with minor alterations.
#
################################################################################
from mudsys import add_cmd
import mud, storage, char, auxiliary, time, string, hooks



# our table of bulletins. Maps name to bulletin
__bulletins__ = { }

# the file we save our bulletin boards to
__bulletin_file__ = "../lib/misc/bulletins"



class Post:
    '''Implements one bulletin board post'''
    def __init__(self, name, time, title, mssg, set = None):
        '''Prepares the post for use'''
        if not set == None:
            self.__poster__ = set.readString("poster")
            self.__time__   = set.readString("time")
            self.__title__  = set.readString("title")
            self.__mssg__   = set.readString("mssg")
        else:
            self.__poster__ = name
            self.__time__   = time
            self.__title__  = title
            self.__mssg__   = mssg

    def store(self):
        '''returns a storage set representation of the post'''
        set = storage.StorageSet()
        set.storeString("poster", self.__poster__)
        set.storeString("time",   self.__time__)
        set.storeString("title",  self.__title__)
        set.storeString("mssg",   self.__mssg__)
        return set

    # various getters and setters for the Post class
    def setPoster(self, name): self.__poster__ = poster
    def setTime(self, time):   self.__time__   = time
    def setTitle(self, title): self.__title__  = title
    def setMssg(self, mssg):   self.__mssg__   = mssg
    def getPoster(self):       return self.__poster__
    def getTime(self):         return self.__time__
    def getTitle(self):        return self.__title__
    def getMssg(self):         return self.__mssg__



class Bulletin:
    '''Implements a bulletin board'''
    def __init__(self, set = None):
        '''Prepares the bulletin board for use'''
        self.__posts__  = []
        self.__groups__ = ""
        self.__rgroups__ = ""
        if not set == None:
            self.__groups__ = set.readString("groups")
            self.__rgroups__ = set.readString("rgroups")
            for post in set.readList("posts").sets():
                self.__posts__.append(Post(None, None, None, None, post))
            
    def store(self):
        '''returns a storage set representation of the board'''
        set   = storage.StorageSet()
        posts = storage.StorageList()
        set.storeString("groups", self.__groups__)
        set.storeString("rgroups", self.__rgroups__)
        set.storeList("posts", posts)
        for post in self.__posts__:
            posts.add(post.store())
        return set

    def post(self, ch, title, mssg):
        '''posts a message onto the bulletin board. Each message is a tuple
           comprised of the poster\'s name, the time of posting, the message
           title, and the message body.'''
        self.__posts__.insert(0, Post(ch.name, time.ctime(), title, mssg))

    def setGroups(self, groups):
        '''sets the groups that have access to this board.'''
        self.__groups__ = groups

    def setRGroups(self, groups):
        '''set the groups which can only read this board.'''
        self.__rgroups__ = groups

    def getGroups(self):
        '''returns the groups that have access to this board.'''
        return self.__groups__

    def getRGroups(self):
        '''returns the groups which can read this board.'''
        return self.__rgroups__

    def getPosts(self):
        '''returns the posts made on the bulletin.'''
        return self.__posts__



################################################################################
# interaction functions
################################################################################
def save_bulletins():
    '''saves all of the bulletin boards to a file'''
    set  = storage.StorageSet()
    list = storage.StorageList()
    set.storeList("list", list)
    for key,val in __bulletins__.iteritems():
        one_set = storage.StorageSet()
        one_set.storeString("key", key)
        one_set.storeSet("val", val.store())
        list.add(one_set)
    set.write(__bulletin_file__)
    set.close()

def load_bulletins():
    set = storage.StorageSet(__bulletin_file__)
    for bulletin in set.readList("list").sets():
        key = bulletin.readString("key")
        __bulletins__[key] = Bulletin(bulletin.readSet("val"))
    set.close()
    return

def do_list_bulletins(ch):
    '''lists all of the bulletins to a character'''
    buf = " %-10s %-28s %-12s %s\r\n" % ("Biuletyn", "Czas", "Autor", "Tytuł")
    buf = buf + "--------------------------------------------------------------------------------\r\n"
    for key in __bulletins__.keys():
        group = __bulletins__[key].getGroups()
        rgroup = __bulletins__[key].getRGroups()
        if group == "" or ch.isInGroup(group) or ch.isInGroup(rgroup):

            if len(__bulletins__[key].getPosts()) == 0:
                buf = buf + " %-10s brak postów\r\n" % key
            else:
                post_one = __bulletins__[key].getPosts()[0]
                buf = (buf + " %-10s %-28s %-12s %s\r\n" %
                       (key, post_one.getTime(), post_one.getPoster(),
                        post_one.getTitle()))
    ch.page(buf)
    return

def do_list_one_bulletin(ch, key):
    '''lists the topics on one bulletin board to a character'''
    if len(__bulletins__[key].getPosts()) == 0:
        ch.send("Biuletyn " + key + " jest obecnie pusty.")
    else:
        i = 1
        buf = " %-3s %-28s %-12s %s\r\n" % ("Num", "Czas", "Autor", "Tytuł")
        buf = buf + "--------------------------------------------------------------------------------\r\n"
        for post in __bulletins__[key].getPosts():
            buf = (buf + " %-3d %-28s %-12s %s\r\n" %
                   (i, post.getTime(), post.getPoster(), post.getTitle()))
            i = i + 1
        ch.page(buf)

def do_post(ch, key, arg):
    '''posts a message to a bulletin board'''
    if not __bulletins__.has_key(key):
        ch.send("Nie istnieje biuletyn o nazwie " + key + ".")
    elif (not __bulletins__[key].getGroups() == "" and
          not ch.isInGroup(__bulletins__[key].getGroups())):
        ch.send("Nie masz uprawnień aby pisać na " + key + ".")
    elif not type(arg) == str:
        ch.send("Musisz podać tytuł dla swojej wiadomości.")
    elif ch.notepad == "":
        ch.send("Twój notatnik jest obecnie pusty. Spróbuj coś napisać.")
    elif ch.hasPrefs('muted'):
        ch.send("Nie możesz pisać nic na forum, ponieważ zostałeś uciszony.")
    else:
        board = __bulletins__[key]
        board.post(ch, arg, ch.notepad)
        save_bulletins()
        ch.send("Wiadomość wysłana.")

def do_read(ch, key, arg):
    '''read a specific messge on a board, or list the board messages'''
    if not __bulletins__.has_key(key):
        ch.send("Nie istnieje biuletyn o nazwie " + key + ".")
        return
    if (not __bulletins__[key].getGroups() == "" and not ch.isInGroup(__bulletins__[key].getGroups())):
        if not __bulletins__[key].getRGroups() == "":
            ch.send("Nie masz uprawnień aby czytać " + key + ".")
            return
        elif not ch.isInGroup(__bulletins__[key].getRGroups()):
            ch.send("Nie masz uprawnień aby czytać " + key + ".")
            return
    if arg == None:
        do_list_one_bulletin(ch, key)
    else:
        try:
            num = string.atoi(arg) - 1
            if num < 0:
                raise IndexError()
            post = __bulletins__[key].getPosts()[num]
            lheader = "Przez " + post.getPoster() + ": " + post.getTitle()
            header  = "%-50s %29s" % (lheader, post.getTime())
            ch.page(header + "\r\n" +
                    "--------------------------------------------------------------------------------\r\n" +
                    post.getMssg())
        except ValueError:
            ch.send("Musisz podać numer wiadomości którą chcesz odczytać.")
        except IndexError:
            ch.send("Wiadomość o tym numerze nie istnieje.")

def do_delete(ch, key, arg):
    '''deleted a post from a board, or an entire board'''
    if not __bulletins__.has_key(key):
        ch.send("Nie istnieje biuletyn o nazwie " + key + ".")
    elif (not __bulletins__[key].getGroups() == "" and
          not ch.isInGroup(__bulletins__[key].getGroups())):
        ch.send("Nie możesz kasować wiadomości z " + key + ".")
    elif ch.isInGroup("admin") and arg == None:
        ch.send("Musisz potwierdzić skasowanie całego biuletynu.")
    elif ch.isInGroup("admin") and arg == "confirm":
        ch.send("Skasowałeś biuletyn, " + key + ".")
        __bulletins__.pop(key)
        save_bulletins()
    else:
        try:
            num = string.atoi(arg) - 1
            if num < 0:
                raise IndexError()
            post = __bulletins__[key].getPosts()[num]
            if post.getPoster() == ch.name or ch.isInGroup("admin"):
                ch.send("Wiadomość skasowana.")
                __bulletins__[key].getPosts().remove(post)
                save_bulletins()
            else:
                ch.send("Nie możesz skasować \"" + post.getTitle() + "\"")
        except ValueError:
            ch.send("Musisz podać numer wiadomości do skasowania.")
        except IndexError:
            ch.send("Wiadomość o tym numerze nie istnieje.")
        


################################################################################
# player commands
################################################################################
def cmd_bulletin(ch, cmd, arg):
    '''The entrypoint into the board system. Allows people to view or post
       messages to bulletin boards. Admin can create/delete boards or messages.

       Syntax:
         biuletyn
         biuletyn czytaj <board> [article number]
         biuletyn pisz <board> <title>
         biuletyn kasuj <board> [article number]
         biuletyn stworz <board> [groups]
       '''
    # if we have no argument, list the board info
    if arg == '':
        do_list_bulletins(ch)
        return

    # try parsing out our arguments
    try:
        subcmd, board, arg = mud.parse_args(ch, True, cmd, arg,
                                            "word word | string")
    except: return
    
    # figure out what subcommand we were trying to do
    subcmd = subcmd.lower()
    board  = board.lower()
    if subcmd == "czytaj":
        do_read(ch, board, arg)
    elif subcmd == "pisz":
        do_post(ch, board, arg)
    elif subcmd == "kasuj":
        do_delete(ch, board, arg)
    elif subcmd == "stworz" and ch.isInGroup("admin"):
        # create a new bulletin board, with group restrictions if neccessary
        if __bulletins__.has_key(board):
            ch.send("Ten biuletyn już istnieje.")
        else:
            __bulletins__[board] = Bulletin()
            ch.send("Nowy biuletyn, " + board + ", został stworzony.")
            if type(arg) == str:
                __bulletins__[board].setGroups(arg)
                ch.send("Grupy biuletynu ustawione na: " + arg)
            save_bulletins()
    elif subcmd == "czytac" and ch.isInGroup("admin"):
        if __bulletins__.has_key(board):
            __bulletins__[board].setRGroups(arg)
            ch.send("Ustawiono uprawnienia do czytania dla grup: " + arg + " na forum " + board)
            save_bulletins()
        else:
            ch.send("Nie istnieje forum o tej nazwie.")
    else:
        # let them know what the valid subcommands are
        buf = "Poprawne komendy to czytaj, pisz, kasuj"
        if ch.isInGroup("admin"): buf = buf + ", stworz, czytac"
        buf = buf + "."
        ch.send(buf)



################################################################################
# hooks
################################################################################
def bulletin_display_hook(info):
    '''displays the bulletin board info to a character when he enters game'''
    ch, = hooks.parse_info(info)
    ch.send(mud.get_motd())
    do_list_bulletins(ch)



################################################################################
# initialization and unloading
################################################################################
load_bulletins()
hooks.add("char_to_game", bulletin_display_hook)

def __unload__():
    '''things that need to be detached when the module is un/reloaded'''
    hooks.remove("char_to_game", bulletin_display_hook)



################################################################################
# add our commands
################################################################################
add_cmd("biuletyn", None, cmd_bulletin, "player", False)
