# -*- coding: utf-8 -*-
###########################################################################
#
# Copyright (C) 2010  Bartek thindil Jasicki
#
# This file is part of Anaria MUD.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
'''
cmd_comm.c

Various commands used in NakedMud(tm) for communicating with other
characters, and NPCs.
'''
import mudsys, mud, inform, hooks, mudsock, history

def can_talk(ch, tgt):
    '''Check if player can talk (is not muted).'''
    if tgt.hasPrefs('muted'):
        ch.send(tgt.name + ' nie odpowie tobie ponieważ jest uciszony.')

def cmd_ask(ch, cmd, arg):
    '''Usage: ask <person> <question>

       This command is used to pose a question to another character. Mostly,
       this is intended to be used to carry on dialogs with NPCs. Ask has a
       local range (i.e. you can only ask questions to people in the same room
       as you.
       '''
    try:
        tgt, question = mud.parse_args(ch, True, cmd, arg,
                                      "ch.room.noself string(pytanie)")
    except: return

    question = question.replace("$", "$$")
    mud.message(ch, tgt, None, None, False, "to_vict",
                "{w$n pyta ciebie, '" + question + "'{n")
    mud.message(ch, tgt, None, None, False, "to_char",
                "{wPytasz $N, '" + question + "'{n")
    can_talk(ch, tgt)
    # run our ask hooks
    hooks.run("ask", hooks.build_info("ch ch str", (ch, tgt, question)))

def cmd_tell(ch, cmd, arg):
    '''Usage: tell <person> <message>

       This command sends a message to another character. Primarily intended
       for player-to-player communication. Players can tell other players
       things even if they are not in the same room.

       see also: reply'''
    try:
        tgt, mssg = mud.parse_args(ch, True, cmd, arg,
                                   "ch.world.noself string(message)")
    except: return

    mssg   = mssg.replace("$", "$$")
    tovict = "{r$n mówi tobie, '" + mssg + "'{n"
    toch   = "{rMówisz $N, '" + mssg + "'{n"
    mud.message(ch, tgt, None, None, False, "to_vict", tovict)
    mud.message(ch, tgt, None, None, False, "to_char", toch)
    history.add_history(ch,   "powiedz", "{r%-10s: %s{n" % (ch.name, mssg))
    history.add_history(tgt,  "powiedz", "{r%-10s: %s{n" % (ch.name, mssg))
    can_talk(ch, tgt)
    hooks.run("tell", hooks.build_info("ch ch str", (ch, tgt, mssg)))

def cmd_chat(ch, cmd, arg):
    '''Usage: chat <message>

       This command will send a message to all players currently logged on.
       '''
    if arg == '':
        ch.send("Powiedzieć co?")
    else:
        arg  = arg.replace("$", "$$")
        mssg = "{y$n gada: '" + arg + "'{n"
        mud.message(ch, None, None, None, False, "to_world", mssg)
        mud.message(ch, None, None, None,False,"to_char", "{yGadasz: '"+arg+"'{n")
        history.add_history(ch, "gadaj", "{y%-10s: %s{n" % (ch.name, arg))

def cmd_wiz(ch, cmd, arg):
    if arg == '':
        ch.send("WizChat co?")
    else:
        mssg = "{b%s WizChats: '{c%s{b'{n" % (ch.name, arg)
        for sock in mudsock.socket_list():
            if sock.ch != None and sock.ch.isInGroup("wizard"):
                sock.ch.send(mssg)

def cmd_say(ch, cmd, arg):
    '''Usage: say <message>

      This command will send a message to everyone in the same room as you. Say,
      like ask, can trigger NPC dialogs.'''
    if arg == '':
        ch.send("Powiedzieć co?")
    else:
        arg = arg.replace("$", "$$")
        mud.message(ch, None, None, None, False, "to_room",
                    "{y$n mówi: '" + arg + "'{n")
        mud.message(ch, None, None, None, False, "to_char",
                    "{yMówisz: '" + arg + "'{n")        

        # run say hooks
        hooks.run("say", hooks.build_info("ch str", (ch, arg)))

def cmd_greet(ch, cmd, arg):
    '''Usage: greet <person>

       NPCs with dialogs will often have something to say when you greet or
       approach then. Greeting an NPC is a way to get them talking.'''
    try:
        tgt, = mud.parse_args(ch, True, cmd, arg, "ch.room.noself")
    except: return

    ch.send("Pozdrawiasz %s$a" % (tgt.name))
    mud.message(ch, tgt, None, None, False, "to_vict", "$n pozdrawia ciebie.")
    ch.sendaround("%s pozdrawia %s$a." % (ch.name, tgt.name))

    # run greet hooks
    hooks.run("greet",      hooks.build_info("ch ch", (ch, tgt)))
    hooks.run("post_greet", hooks.build_info("ch ch", (ch, tgt)))

def cmd_emote(ch, cmd, arg):
    '''Usage: emote <text>

       Send a special text message to the room you are in. The message is
       preceded by your name, unless you put a $n somewhere in the text, in
       which case the $n is replaced by your name. For example:

       > emote A gunshot sounds, and $n is laying on the ground, dead.

       Would show a message to everyone in the room saying that you are dead
       to a gunshot.'''
    if arg == '':
        ch.send("Co chcesz odegrać?")
    else:
        # see if a $n is within the argument ... if there is, let the person
        # put his or her name where it's wanted. Otherwise, tag it onto the
        # front of the message
        if arg.find("$n") == -1:
            arg = "$n " + arg
        mud.message(ch, None, None, None, False, "to_room, to_char", arg)

def cmd_gemote(ch, cmd, arg):
    '''Gemote is similar to emote, except that it sends a mud-wide message
       instead of a room-specific message.'''
    if arg == '':
        ch.send("Co chcesz odegrać?")
    else:
        # same as emote, but global
        if arg.find("$n") == -1:
            arg = "$n " + arg
        mud.message(ch, None, None, None, False, "to_world, to_char",
                    "{bGLOBAL:{c " + arg + "{n")

def cmd_page(ch, cmd, arg):
    '''Usage: page <person> <message>

       Paging a person will send them a message, as well as making a beeping
       sound on their computer to get their attention. Page can be used on
       anyone in the mud, regardless if you are in the same room as them or not.
       '''
    try:
        tgt, mssg = mud.parse_args(ch, True, cmd, arg,
                                   "ch.world.noself string(message)")
    except: return
    ch.send("\007\007Wysłałeś gońca do " + ch.see_as(tgt) + "$g")
    tgt.send("\007\007*" + tgt.see_as(ch) + "* " + mssg)
    can_talk(ch, tgt)

def cmd_reply(ch, cmd, arg):
    '''Reply to last private message teller.'''
    if not ch.hasvar('tell'):
        ch.send("Nikt jeszcze nie powiedział nic tobie.")
        return
    try:
        tgt, = mud.parse_args(ch, True, cmd, ch.getvar('tell'), 'ch.world.noself')
    except:
        return
    if len(arg) == 0:
        ch.send("Co chcesz powiedzieć?")
        return
    ch.act("powiedz " + tgt.name + " " + arg)

################################################################################
# game hook
################################################################################

def reply_hook(info):
    '''Remove reply variable from character.'''
    ch, = hooks.parse_info(info)
    if ch.hasvar('tell'):
        ch.delvar('tell')

################################################################################
# add our commands
################################################################################
mudsys.add_cmd("pytaj",     None, cmd_ask,   "player", False)
mudsys.add_cmd("mow",     None, cmd_say,   "player", False)
mudsys.add_cmd("'",       None, cmd_say,   "player", False)
mudsys.add_cmd("powiedz",    None, cmd_tell,  "player", False)
mudsys.add_cmd("gadaj",    None, cmd_chat,  "player", False)
mudsys.add_cmd("wizchat", "wiz",cmd_wiz,   "wizard", False)
mudsys.add_cmd("\"",      None, cmd_chat,  "player", False)
mudsys.add_cmd("goniec",    None, cmd_page,  "player", False)
mudsys.add_cmd("pozdrow",   None, cmd_greet, "player", False)
mudsys.add_cmd("odegraj",   None, cmd_emote, "player", False)
mudsys.add_cmd("godegraj",  None, cmd_gemote,"player", False)
mudsys.add_cmd(":",       None, cmd_emote, "player", False)
mudsys.add_cmd("odpowiedz", None, cmd_reply, "player", False)

mudsys.create_bit('char_prfs', 'muted')

def chk_room_communication(ch, cmd):
    if ch.hasPrefs('muted'):
        ch.send('Nie możesz mówić ponieważ zostałeś uciszony.')
        return False
    if ch.pos == 'sleeping':
        pos = 'śpisz'
    if ch.pos == 'unconscious':
        pos = 'jesteś nieprzytomny'
    if ch.pos in ("sleeping", "unconscious"):
        ch.send("Nie możesz tego zrobić kiedy " + pos + ".")
        return False

for cmd in ["pytaj", "mow", "'", "pozdrow", "odegraj", ":", "odpowiedz"]:
    mudsys.add_cmd_check(cmd, chk_room_communication)

# register our history handling
history.register_comm_history("gadaj", lambda ch: "gadaj")
history.register_comm_history("powiedz", lambda ch: ch.name)

hooks.add("char_from_game", reply_hook)

def __unload__():
    '''remove hooks for communications.'''
    hooks.remove('char_from_game', reply_hook)
