# -*- coding: utf-8 -*-
###########################################################################
#
# Copyright (C) 2010  Bartek thindil Jasicki
#
# This file is part of Anaria.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
'''
weapons.py

Weapons item type.
'''
import olc, mudsys, auxiliary, mudsock, storage, mud, utils, hooks



################################################################################
# item type definition
################################################################################

# a map from our type to possible subtypes.
# Also the default subtype when we create an item of the specified type.
wtype_map  = { }
wtype_dflt = { }

class WeaponData:
    '''Data for weapons - type, subtype, damage and skill'''

    __item_type__ = "weapon"

    def __init__(self, set = None):
        self.type = "melee"
        self.subtype = "crush"
        self.damage = 0
        self.skill = "Walka wręcz"
        self.prepare = 1

        if set != None:
            self.type = set.readString('type')
            self.subtype = set.readString('subtype')
            self.damage = set.readInt('damage')
            self.skill = set.readString('skill')
            if set.contains('prepare'):
                self.prepare = set.readInt('prepare')

    def copyTo(self, to):
        to.type = self.type
        to.subtype = self.type
        to.damage = self.damage
        to.skill = self.skill
        to.material = self.material
        to.prepare = self.prepare

    def copy(self):
        newData = WeaponData()
        self.copyTo(newData)
        return newData

    def store(self):
        set = storage.StorageSet()
        set.storeString('type', self.type)
        set.storeString('subtype', self.subtype)
        set.storeInt('damage', self.damage)
        set.storeString('skill', self.skill)
        set.storeString('material', self.material)
        set.storeInt('prepare', self.prepare)
        return set

################################################################################
# creating and interacting with weapon types
################################################################################
def register_weapon_type(type, dflt_subtype = "unknown"):
    '''create a new weapon type, and add a default subtype'''
    subtypes = [dflt_subtype]
    wtype_map[type]  = subtypes
    wtype_dflt[type] = dflt_subtype

def register_weapon_subtype(type, subtype):
    '''create a new subtype for the specified weapon type'''
    if wtype_map.has_key(type):
        subtype  = subtype.lower()
        subtypes = wtype_map[type]
        if not subtype in subtypes:
            subtypes.append(subtype)

def get_weapon_dflt(type):
    '''get the default subtype for a weapon'''
    if wtype_dflt.has_key(type):
        return wtype_dflt[type]
    return None

def get_weapon_types():
    '''list all weapons types'''
    return wtype_map.keys()

def get_weapon_subtypes(type):
    '''list all subtypes for a weapon'''
    if wtype_map.has_key(type):
        return [t for t in wtype_map[type]]
    return None

def is_weapon_type(type):
    '''return whether type is a registered weapon type'''
    return wtype_map.has_key(type)

def is_weapon_subtype(type, subtype):
    '''return whether subtype is a registered subtype for type'''
    if wtype_map.has_key(type):
        return subtype in wtype_map[type]
    return False
    


################################################################################
# Weapon OLC
################################################################################
__WEAPOLC_TYPE__     = 1
__WEAPOLC_SUBTYPE__  = 2
__WEAPOLC_DAMAGE__   = 3
__WEAPOLC_SKILL__    = 4
__WEAPOLC_PREPARE__  = 5

weapolc_opt_map = {
    '1' : (__WEAPOLC_TYPE__,     "Wybierz typ: "),
    '2' : (__WEAPOLC_SUBTYPE__,  "Wybierz podtyp: "),
    '3' : (__WEAPOLC_DAMAGE__, "Wprowadź zadawane obrażenia: "),
    '4' : (__WEAPOLC_SKILL__,     "Wprowadź umięjetność: "),
    '5' : (__WEAPOLC_PREPARE__, "Wprowadź opóźnienie: ")
    }

def weapolc_menu(sock, data):
    '''display our menu interface'''
    sock.send("{g1) Typ         : {c" + data.type    + "\n" + \
              "{g2) Podtyp      : {c" + data.subtype + "\n" + \
              "{g3) Obrażenia   : {c" + str(data.damage) + "\n" + \
              "{g4) Umiejętność : {c" + data.skill + "\n" + \
              "{g5) Opóźnienie  : {c" + str(data.prepare))

def weapolc_chooser(sock, data, option):
    '''figure out what field we want to edit, and display a prompt'''
    if not weapolc_opt_map.has_key(option):
        return olc.MENU_CHOICE_INVALID
    ret, mssg = weapolc_opt_map[option]

    # list our type and subtype options
    opts = None
    if ret == __WEAPOLC_TYPE__:
        opts = get_weapon_types()
    elif ret == __WEAPOLC_SUBTYPE__:
        opts = get_weapon_subtypes(data.type)

    # do we have to list subtype options?
    if opts != None:
        utils.olc_display_table(sock, opts, 4)
    sock.send_raw(mssg)
    return ret

def weapolc_parser(sock, data, choice, arg):
    '''parse out what we want to change a specified field to'''
    if choice != __WEAPOLC_SKILL__:
        try:
            arg = int(arg)
        except: return False
    
    if choice == __WEAPOLC_TYPE__:
        opts = get_weapon_types()
        arg  = int(arg)
        if arg < 0 or arg >= len(opts):
            return False

        newtype = opts[arg]
        if newtype != data.type:
            data.damage = 1
            data.skill = 'Miecze'
            data.subtype = get_weapon_dflt(newtype)
            data.type = newtype

    elif choice == __WEAPOLC_SUBTYPE__:
        opts = get_weapon_subtypes(data.type)
        arg  = int(arg)
        if arg < 0 or arg >= len(opts):
            return False
        data.subtype = opts[arg]

    elif choice == __WEAPOLC_DAMAGE__:
        data.damage = max(int(arg), 1)
    elif choice == __WEAPOLC_SKILL__:
        data.skill  = str(arg)
    elif choice == __WEAPOLC_PREPARE__:
        data.prepare = max(int(arg), 1)
    else:
        return False
    return True

def weapolc_to_proto(data):
    '''return a script that will generate relevant info to create this item'''
    buf  = 'me.get_type_data("weapon").type = "' + data.type + '"\n'
    buf += 'me.get_type_data("weapon").subtype = "' + data.subtype + '"\n'
    buf += 'me.get_type_data("weapon").damage = ' + str(data.damage) + '\n'
    buf += 'me.get_type_data("weapon").skill = "' + data.skill + '"\n'
    buf += 'me.get_type_data("weapon").prepare = ' + str(data.prepare) + '\n'
    return buf

################################################################################
# game hooks
################################################################################
def weapon_desc_info(info):
    '''appends weapon information about an object to the description
       buffer when a person looks at it.'''
    obj, ch = hooks.parse_info(info)

    # no point in continuing
    if not obj.istype("weapon"):
        return

    if obj.get_type_data("weapon").type == "melee":
        wtype = "do walki wręcz"
    if obj.get_type_data("weapon").subtype == "cut":
        dtype = "tnące"
    elif obj.get_type_data("weapon").subtype == "impaling":
        dtype = "przebijające"
    elif obj.get_type_data("weapon").subtype == "crush":
        dtype = "miażdżone"
    if obj.get_type_data("weapon").prepare == 1:
        preparation = "rundę."
    else:
        preparation = str(obj.get_type_data("weapon").prepare) + " rund(y)."

    desc = " Jest to broń " + wtype + ", zadająca rany " + dtype + ". Zwiększa zadawane obrażenia o " + str(obj.get_type_data("weapon").damage) + "%. Wymaga umiejętności: " + obj.get_type_data("weapon").skill + ". Można nią atakować co " + preparation

    # append our description
    ch.look_buf += desc

################################################################################
# initialization
################################################################################

# initialize our item data, and its OLC
mudsys.item_add_type(WeaponData.__item_type__, WeaponData)
olc.item_add_olc(WeaponData.__item_type__,
                 weapolc_menu, weapolc_chooser, weapolc_parser,
                 None, weapolc_to_proto)

# display hooks for weapon
hooks.add("append_obj_desc", weapon_desc_info)

# initialize some default weapons types and subtypes
register_weapon_type("melee",    "cut")
register_weapon_subtype("melee", "impaling")
register_weapon_subtype("melee", "crush")

# add worn type for weapons
mudsys.add_worn_type('ohweapon', 'dzierżone')
mudsys.add_worn_type('thweapon', 'dzierżone, trzymane')

