/*
 * Copyright (C), 2010 Bartek thindil Jasicki
 *
 * This file is part of Anaria.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
//*****************************************************************************
//
// pyzone.c
//
// Contains a python zone module, and a Zone class that is a python 
// wrapper for Anaria zones.
//
//*****************************************************************************

#include "../mud.h"
#include "../utils.h"
#include "../zone.h"
#include "../world.h"

#include "scripts.h"
#include "pyplugs.h"
#include "pyauxiliary.h"
#include "pyzone.h"

//*****************************************************************************
// local structures and defines
//*****************************************************************************

// a list of the get/setters on the Zone class
LIST *pyzone_getsetters = NULL;

// a list of the methods on the Zone class
LIST *pyzone_methods = NULL;

typedef struct 
{
  PyObject_HEAD
  ZONE_DATA *zone;
} PyZone;

//*****************************************************************************
// allocation, deallocation, initialization, and comparison
//*****************************************************************************
void PyZone_dealloc (PyZone *self) 
{
  self->ob_type->tp_free((PyObject*)self);
}

PyObject *PyZone_new(PyTypeObject *type, PyObject *args, PyObject *kwds) 
{
    PyZone *self;
    self = (PyZone *)type->tp_alloc(type, 0);
    self->zone = NULL;
    return (PyObject *)self;
}

int PyZone_init(PyZone *self, PyObject *args, PyObject *kwds) 
{
  char *kwlist[] = {"key", NULL};
  char *key = NULL;

  // get the uid
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "s", kwlist, &key)) {
    PyErr_Format(PyExc_TypeError, "Zones may only be created using a key");
    return -1;
  }

  self->zone = zoneLoad(gameworld, key);
  return 0;
}

int PyZone_compare(PyZone *zone1, PyZone *zone2) 
{
  return strcasecmp(zoneGetName(zone1->zone), zoneGetName(zone2->zone));
}

//*****************************************************************************
// getters and setters for the Zone class
//*****************************************************************************
PyObject *PyZone_getname(PyZone *self, void *closure) 
{
  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  if (zone != NULL)
    { 
      return Py_BuildValue("s", zoneGetName(zone));
    }
  else
    {
      return NULL;
    }
}

PyObject *PyZone_getdesc(PyZone *self, void *closure) 
{
  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  if (zone != NULL)
    { 
      return Py_BuildValue("s", zoneGetDesc(zone));
    }
  else
    {
      return NULL;
    }
}

PyObject *PyZone_geteditors(PyZone *self, void *closure) 
{
  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  if (zone != NULL)
    { 
      return Py_BuildValue("s", zoneGetEditors(zone));
    }
  else
    {
      return NULL;
    }
}

PyObject *PyZone_getkey(PyZone *self, void *closure) 
{
  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  if (zone != NULL)
    { 
      return Py_BuildValue("s", zoneGetKey(zone));
    }
  else
    {
      return NULL;
    }
}

int PyZone_setname(PyZone *self, PyObject *value, void *closure) 
{
  if (value == NULL) 
    {
      PyErr_Format(PyExc_TypeError, "Cannot delete zone's name");
      return -1;
    }
  
  if (!PyString_Check(value)) 
    {
      PyErr_Format(PyExc_TypeError, "Zone names must be strings");
      return -1;
    }

  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  zoneSetName(zone, PyString_AsString(value));
  return 0;
}

int PyZone_setdesc(PyZone *self, PyObject *value, void *closure) 
{
  if (value == NULL) 
    {
      PyErr_Format(PyExc_TypeError, "Cannot delete zone's description");
      return -1;
    }
  
  if (!PyString_Check(value)) 
    {
      PyErr_Format(PyExc_TypeError, "Zone description must be strings");
      return -1;
    }

  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  zoneSetDesc(zone, PyString_AsString(value));
  return 0;
}

int PyZone_seteditors(PyZone *self, PyObject *value, void *closure) 
{
  if (value == NULL) 
    {
      PyErr_Format(PyExc_TypeError, "Cannot delete zone's editors");
      return -1;
    }
  
  if (!PyString_Check(value)) 
    {
      PyErr_Format(PyExc_TypeError, "Zone editors must be strings");
      return -1;
    }

  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  zoneSetEditors(zone, PyString_AsString(value));
  return 0;
}

int PyZone_setkey(PyZone *self, PyObject *value, void *closure) 
{
  if (value == NULL) 
    {
      PyErr_Format(PyExc_TypeError, "Cannot delete zone's key");
      return -1;
    }
  
  if (!PyString_Check(value)) 
    {
      PyErr_Format(PyExc_TypeError, "Zone key must be strings");
      return -1;
    }

  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  zoneSetKey(zone, PyString_AsString(value));
  return 0;
}

//*****************************************************************************
// methods for the Zone class
//*****************************************************************************

// Load selected zone
PyObject *PyZone_load_zone(PyZone *self, PyObject *args)
{
  char *keyword = NULL;
  if (!PyArg_ParseTuple(args, "s", &keyword)) 
    {
      PyErr_Format(PyExc_TypeError,
		   "You must specify zone key for zone to load.");
      return NULL;
    }

  ZONE_DATA *zone = zoneLoad(gameworld, keyword);
  PyObject *retval = Py_BuildValue("O", zone);
  return retval;
}

//
// returns the specified piece of auxiliary data from the account
// if it is a piece of python auxiliary data.
PyObject *PyZone_get_auxiliary(PyZone *self, PyObject *args) 
{
  char *keyword = NULL;
  if (!PyArg_ParseTuple(args, "s", &keyword)) 
    {
      PyErr_Format(PyExc_TypeError,
		   "getAuxiliary() must be supplied with the name that the "
		   "auxiliary data was installed under!");
      return NULL;
    }

  // make sure we exist
  ZONE_DATA *zone = PyZone_AsZone((PyObject *)self);
  if (zone == NULL) 
    {
      PyErr_Format(PyExc_StandardError,
		   "Tried to get auxiliary data for a nonexistant zone.");
      return NULL;
    }

  // make sure the auxiliary data exists
  if (!pyAuxiliaryDataExists(keyword)) 
    {
      PyErr_Format(PyExc_StandardError,
		   "No auxiliary data named '%s' exists!", keyword);
      return NULL;
    }

  PyObject *data = zoneGetAuxiliaryData(zone, keyword);
  if (data == NULL)
    {
      data = Py_None;
    }
  PyObject *retval = Py_BuildValue("O", data);
  return retval;
}

//*****************************************************************************
// structures to define our methods and classes
//*****************************************************************************
PyTypeObject PyZone_Type = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "zone.Zone",               /*tp_name*/
    sizeof(PyZone),            /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)PyZone_dealloc,/*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    (cmpfunc)PyZone_compare,   /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Python Zone object",      /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    0,                         /* tp_methods */
    0,                         /* tp_members */
    0,                         /* tp_getset */ 
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)PyZone_init,     /* tp_init */
    0,                         /* tp_alloc */
    PyZone_new,                /* tp_new */
};

PyMethodDef zone_module_methods[] = {
  {NULL, NULL, 0, NULL}  /* Sentinel */
};

//*****************************************************************************
// implementation of pyzone.h
//*****************************************************************************
void PyZone_addGetSetter(const char *name, void *g, void *s,const char *doc)
{
  // make sure our list of get/setters is created
  if (pyzone_getsetters == NULL)
    {
      pyzone_getsetters = newList();
    }

  // make the GetSetter def
  PyGetSetDef *def = calloc(1, sizeof(PyGetSetDef));
  def->name        = strdup(name);
  def->get         = (getter)g;
  def->set         = (setter)s;
  def->doc         = (doc ? strdup(doc) : NULL);
  def->closure     = NULL;
  listPut(pyzone_getsetters, def);
}

void PyZone_addMethod(const char *name, void *f, int flags, const char *doc)
{
  // make sure our list of methods is created
  if(pyzone_methods == NULL)
    { 
      pyzone_methods = newList();
    }

  // make the Method def
  PyMethodDef *def = calloc(1, sizeof(PyMethodDef));
  def->ml_name     = strdup(name);
  def->ml_meth     = (PyCFunction)f;
  def->ml_flags    = flags;
  def->ml_doc      = (doc ? strdup(doc) : NULL);
  listPut(pyzone_methods, def);
}

// initialize zone for use. This must be called AFTER 
PyMODINIT_FUNC
init_PyZone(void) 
{
    PyObject* module = NULL;

    // add all of the basic getsetters
    PyZone_addGetSetter("name", PyZone_getname, PyZone_setname, 
			   "The zones's name.");
    PyZone_addGetSetter("desc", PyZone_getdesc, PyZone_setdesc, 
			   "The zones's description.");
    PyZone_addGetSetter("editors", PyZone_geteditors, PyZone_seteditors, 
			   "The zones's editors.");
    PyZone_addGetSetter("key", PyZone_getkey, PyZone_setkey, 
			   "The zones's key.");

    // add all of the basic methods
    PyZone_addMethod("getAuxiliary", PyZone_get_auxiliary, METH_VARARGS,
      "getAuxiliary(name)\n"
      "\n"
      "Returns zone's auxiliary data of the specified name.");
    PyZone_addMethod("aux", PyZone_get_auxiliary, METH_VARARGS,
      "Alias for zone.Zone.getAuxiliary.");
    PyZone_addMethod("load", PyZone_load_zone, METH_VARARGS,
      "Load selected zone.");

    // add in all the getsetters and methods
    makePyType(&PyZone_Type, pyzone_getsetters, pyzone_methods);
    deleteListWith(pyzone_getsetters, free); 
    pyzone_getsetters = NULL;
    deleteListWith(pyzone_methods, free); 
    pyzone_methods = NULL;

    // make sure the zone class is ready to be made
    if (PyType_Ready(&PyZone_Type) < 0)
        return;

    // initialize the module
    module = Py_InitModule3("zone", zone_module_methods,
			    "Contains the Python wrapper for zones.");

    // make sure the module parsed OK
    if (module == NULL)
      return;

    // add the Zone class to the account module
    PyTypeObject *type = &PyZone_Type;
    PyModule_AddObject(module, "Zone", (PyObject *)type);
    Py_INCREF(&PyZone_Type);
}

ZONE_DATA *PyZone_AsZone(PyObject *zone) 
{
  return ((PyZone *)zone)->zone;
}

int PyZone_Check(PyObject *value) 
{
  return PyObject_TypeCheck(value, &PyZone_Type);
}

PyObject *newPyZone(ZONE_DATA *zone) 
{
  PyZone *py_zone = (PyZone *)PyZone_new(&PyZone_Type, NULL, NULL);
  py_zone->zone = zone;
  return (PyObject *)py_zone;
}
